/***********************************************************************
 * ME_Array.c
 *
 * Generate a RAM-array which can used for performance analysis
 *
 ***********************************************************************
 * License
 * (C) Copyright 2019 Falk Richter, falk.richter at yandex.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 ***********************************************************************/ 

#include <limits.h>

#include "ME_Array.h"
#include "EX_Error.h"


/* -------------------------------------------------------------------- */ 
/* the calculation of the stimulus for a width of one byte */
/* -------------------------------------------------------------------- */ 
void ME_vGenUint8Array(uint32_T u32Size, uint8_T * pu8Data) {
  bool_T bRise = TRUE;
  uint8_T u8Value = 0;
  uint32_T u32i;

  /* we fill the array with a triangle signal */
  for (u32i = 0; u32i < u32Size; u32i++) { 
    /* set the value into the array */
    pu8Data[u32i] = u8Value;

    /* check if we on the rising edge */
    if (bRise) {
      if (u8Value < UCHAR_MAX) {
	/* prepare the next value */
	u8Value++; 
      } else {
	/* the value is 255, we are on the positive peak */
	/* we change the direction */
	bRise = FALSE;
	/* and set the next value with 254 */
	u8Value--;
      }
    } else {
      /* we are on the falling edge */
      if (u8Value > 0) {
	/* prepare the next value */
	u8Value--;
      } else {
	/* the value is zero, we are on the negative peak */
	/* we change the direction */
	bRise = TRUE;
	/* and set the next value with 1 */
	u8Value++;
      }
    }
  } 
}


/* -------------------------------------------------------------------- */ 
/* the calculation of the stimulus for a width of four bytes */
/* -------------------------------------------------------------------- */ 
void ME_vGenUint32Array(uint32_T u32Size, uint32_T * pu32Data) {
  bool_T bRise = TRUE;
  uint32_T u32Value = 0; 
  uint32_T u32i;

  /* we fill the array with a triangle signal */
  for (u32i = 0; u32i < u32Size; u32i++) { 
    /* set the value into the array */
    pu32Data[u32i] = u32Value;

    /* check if we on the rising edge */
    if (bRise) {
      if (u32Value < UINT_MAX) {
	/* prepare the next value */
	u32Value++; 
      } else {
	/* the value is UINT_MAX, we are on the positive peak */
	/* we change the direction */
	bRise = FALSE;
	/* and set the next value with UINT_MAX - 1 */
	u32Value--;
      }
    } else {
      /* we are on the falling edge */
      if (u32Value > 0) {
	/* prepare the next value */
	u32Value--;
      } else {
	/* the value is zero, we are on the negative peak */
	/* we change the direction */
	bRise = TRUE;
	/* and set the next value with 1 */
	u32Value++; 
      }
    }
  } 
}


/* -------------------------------------------------------------------- */ 
/* copy two arrays ( data type: one byte ) */
/* -------------------------------------------------------------------- */ 
void ME_vCopyUint8Array(uint32_T u32Size, uint8_T * pu8Source, uint8_T * pu8Destination) { 
  uint32_T u32i;

  for (u32i = 0; u32i < u32Size; u32i++) {
	/* put source into the destination array */
	pu8Destination[u32i] = pu8Source[u32i]; 
  } 
}


/* -------------------------------------------------------------------- */ 
/* copy two arrays ( data type: four bytes ) */
/* -------------------------------------------------------------------- */ 
void ME_vCopyUint32Array(uint32_T u32Size, uint32_T * pu32Source, uint32_T * pu32Destination) { 
  uint32_T u32i;

  for (u32i = 0; u32i < u32Size; u32i++) {
    /* put source into the destination array */
    pu32Destination[u32i] = pu32Source[u32i];
  } 
} 

