/***********************************************************************
 * ME_Measure.c
 *
 * Prepare and execute the different measurements
 *
 ***********************************************************************
 * License
 * (C) Copyright 2019 Falk Richter, falk.richter at yandex.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 ***********************************************************************/ 
#include <stdlib.h> /* for malloc */
#include <stdio.h>  /* for fprintf */
/*
#include <limits.h>
*/
#include <fcntl.h>  /* for file access */
#include <unistd.h> /* for file access */

#include "ME_Time.h"
#include "ME_Array.h"
#include "EX_Error.h"
#include "PRJ_Settings.h"
#include "ME_Measure.h"
#include "ME_LogFile.h"

static bool_T mg_bDoTest1;
static bool_T mg_bDoTest2;
static bool_T mg_bDoTest3;

static uint32_T mg_u32MeasRepetitions;
static uint32_T mg_u32ArraySizeCacheTest;
static uint32_T mg_u32ArraySizeMemTest; 
static uint32_T mg_u32FileSize;

static void vCmdLineOut(char * pszMessage, uint64_T u64TimeDiff);
static void vCmdLineOutArray(char * pszMessage, uint64_T * pu64TimeDiffs, uint32_T u32CntOfBytes); 
static uint64_T u64MemoryCopy(uint32_T u32ArraySize);
static uint64_T u64WriteFile(uint32_T u32ArraySize, char * pszFileName);
static uint64_T u64ReadFile(char * pszFileName);

/* -------------------------------------------------------------------- */ 
/* initialize the measurement module */
/* -------------------------------------------------------------------- */ 
void ME_vMeasInit(void) {

  /* we calculate the size of the bytes for the cache and memory
   * test depending on the cache size */
  uint32_T u32CacheSize; 
  /* parse here the type of testes once */
  uint32_T u32TestTypes = PRJxSET_u32GetTestTypes();

  if ( u32TestTypes & 1u ) {
    mg_bDoTest1 = TRUE;
  } else {
    mg_bDoTest1 = FALSE; 
  }

  if ( u32TestTypes & 2u ) {
    mg_bDoTest2 = TRUE;
  } else {
    mg_bDoTest2 = FALSE; 
  }

  if ( u32TestTypes & 4u ) {
    mg_bDoTest3 = TRUE;
  } else {
    mg_bDoTest3 = FALSE; 
  } 

  /* count of repetitions for the cache and memory test to have 
   * a feeling for the deviations */
  mg_u32MeasRepetitions = 10u;

  /* set the exact value for the cache size */ 
  switch ( PRJxSET_u32GetCacheSize() ) {
    case 64u:
      {
	/* 64k bytes cache size */ 
	u32CacheSize = 65536u;
      }
      break;
    case 128u:
      {
	/* 128k bytes cache size */
	u32CacheSize = 131072u;
      }
      break;
    case 512u:
      {
	/* 512k bytes cache size */
	u32CacheSize = 524288u;
      }
      break;
    case 1024u:
      {
	/* 1M bytes cache size */
	u32CacheSize = 1048576u;
      }
      break;    
    default: 
    case 256u:
      {
	/* 256k bytes cache size */
	u32CacheSize = 262144u;
      }
      break; 
  }

  /* The array size for the cache test is a quarter of the cache size 
   * to ensure the test fits into the cache.
   * The array size for the memory test is four times cache size to
   * ensure the test doesn't fit into the cache... :o) 
   *
   * We use an integer data type with 32 bits (four bytes) because 
   * this is the native bus width of the 486 cpu. So at each
   * operation we transfer four bytes: we consider this already in
   * the factors (e.g.): 
   * normally: mg_u32ArraySizeCacheTest = 65536u / 4u; 
   * with four bytes / operation: mg_u32ArraySizeCacheTest = 65536u / 4u / 4u;
   * normally: mg_u32ArraySizeMemTest   = 65536u * 4u;
   * with four bytes / operation: mg_u32ArraySizeMemTest   = 65536u * 4u / 4u;
   * */ 
  mg_u32ArraySizeCacheTest = u32CacheSize / 16u;
  mg_u32ArraySizeMemTest   = u32CacheSize;

  /* file size in bytes for the file write / read test */
  mg_u32FileSize = 4194304;

  /* initialize the timer module */
  ME_vTimerInit();
}


/* -------------------------------------------------------------------- */ 
/* Start all measurements */
/* -------------------------------------------------------------------- */ 
bool_T ME_bMeasStart(void) {

  /* time measurement result */
  uint64_T u64TimeDiff;
  /* loop index */
  uint32_T u32Idx;
  /* string array for the output */
  char aszOutput[80u];
 /* count of copied bytes */
 uint32_T u32BytesCopied; 

  /* measurement result array */
  uint64_T * pu64TimeDiffs = (uint64_T *) malloc( mg_u32MeasRepetitions * sizeof(uint64_T) );
  /* check if the allocation failed */
  if ( pu64TimeDiffs == NULL ) {
    /* we through an exception */
    char aszErrorMsg[] = "Malloc failed for pu64TimeDiffs"; 
    EX_vSetError(aszErrorMsg, __FILE__, __LINE__);
    return(FALSE); 
  } 

  /* the memory block copy test */
  if ( mg_bDoTest1 ) { 
    fprintf(stdout, "Start the memory test\n");
    fprintf(stdout, "---------------------\n"); 	

    /* start the memory copy also ten times */
    for (u32Idx = 0; u32Idx < mg_u32MeasRepetitions; u32Idx++) {
      *pu64TimeDiffs++ = u64MemoryCopy(mg_u32ArraySizeMemTest);
    }
    /* move back the pointer */
    pu64TimeDiffs -= mg_u32MeasRepetitions;

    /* Calculate the amount of bytes which we have copied (4 * ArraySize because we wrote 4 bytes per sample) */
    u32BytesCopied = mg_u32ArraySizeMemTest * 4u;
 
    /* Prepare the string output */
    snprintf(aszOutput, 80u, "# Memory test: copy of an array with %d bytes needs:", u32BytesCopied); 
    
    /* calculate and put out the result */
    vCmdLineOutArray(aszOutput, pu64TimeDiffs, u32BytesCopied);
  }
  
  /* the cache block copy test */
  if (mg_bDoTest2) { 
    fprintf(stdout, "Start the cache test\n");
    fprintf(stdout, "--------------------\n"); 
    
    /* start the memory copy also ten times */
    for (u32Idx = 0; u32Idx < mg_u32MeasRepetitions; u32Idx++) {
      *pu64TimeDiffs++ = u64MemoryCopy(mg_u32ArraySizeCacheTest);
    }
    /* move back the pointer */
    pu64TimeDiffs -= mg_u32MeasRepetitions; 

    /* Calculate the amount of bytes which we have copied (4 * ArraySize because we wrote 4 bytes per sample) */
    u32BytesCopied = mg_u32ArraySizeCacheTest * 4u;
 
    /* Prepare the string output */
    snprintf(aszOutput, 80u, "# Cache test: copy of an array with %d bytes needs:", u32BytesCopied); 
    
    /* calculate and put out the result */
    vCmdLineOutArray(aszOutput, pu64TimeDiffs, u32BytesCopied);
  }
  
  /* file write and read test */
  if (mg_bDoTest3) { 
    fprintf(stdout, "Start the file write and read test\n");
    fprintf(stdout, "----------------------------------\n");

    /* start the file writing */
    u64TimeDiff = u64WriteFile(mg_u32FileSize, PRJxSET_pszGetFileName()); 
    
    /* Prepare the string output */
    snprintf(aszOutput, 80u, "# Writing a text file with %d bytes needs:", mg_u32FileSize); 
    
    /* put out the result */
    vCmdLineOut(aszOutput, u64TimeDiff);

    /* now we read back the file */
    u64TimeDiff = u64ReadFile(PRJxSET_pszGetFileName());
    
    /* put out the result */
    vCmdLineOut("# Read back the same file needs:", u64TimeDiff); 
  }
  return(TRUE);
} 


/* -------------------------------------------------------------------- */ 
/* allocate and copy the memory block */
/* -------------------------------------------------------------------- */ 
static uint64_T u64MemoryCopy(uint32_T u32ArraySize) { 
  
  /* the data source array */
  uint32_T * pu32DataSource; 

  /* the data destination array */
  uint32_T * pu32DataDest; 

  if ( PRJxSET_bGetVerboseMode() ) {
    fprintf(stdout, "Ask for source and destination memory.\n");
  }

  /* allocate memory for the source array */
  pu32DataSource = (uint32_T *) malloc( u32ArraySize * sizeof(uint32_T) );
  
  /* check if the allocation failed */
  if ( pu32DataSource == NULL ) {
    /* we through an exception */
    char aszErrorMsg[] = "Can't allocate enough memory for pu32DataSource"; 
    EX_vSetError(aszErrorMsg, __FILE__, __LINE__);
    return(FALSE);
  } 

  /* allocate memory for the destination array */
  pu32DataDest = (uint32_T *) malloc( u32ArraySize * sizeof(uint32_T) );
  
  /* check if the allocation failed */
  if ( pu32DataDest == NULL ) {
    /* we through an exception */
    char aszErrorMsg[] = "Can't allocate enough memory for pu32DataDest"; 
    EX_vSetError(aszErrorMsg, __FILE__, __LINE__);
    return(FALSE);
  } 

  if ( PRJxSET_bGetVerboseMode() ) {
    fprintf(stdout, "Fill the source array with a saw tooth.\n");
  } 
  
  /* fill the source array */
  ME_vGenUint32Array(u32ArraySize, pu32DataSource); 

  if ( PRJxSET_bGetVerboseMode() ) {
    fprintf(stdout, "Start the copy process.\n");
  }
  
  /* start the time measurement */
  ME_vStartTimer();

  /* copy the memory block */
  ME_vCopyUint32Array(u32ArraySize, pu32DataSource, pu32DataDest);

  /* stop the time measurement */
  ME_vStopTimer();

  if ( PRJxSET_bGetVerboseMode() ) {
    fprintf(stdout, "Give source and destination memory free.\n");
  }
  
  /* free the allocated memory again */ 
  free(pu32DataSource); 
  free(pu32DataDest); 

  return( ME_u64GetTimeDiff() );
}


/* -------------------------------------------------------------------- */ 
/* Write a file to the hard disc */
/* -------------------------------------------------------------------- */ 
static uint64_T u64WriteFile(uint32_T u32ArraySize, char * pszFileName) {

  /* the data source array */
  char * pcDataSource;
  /* file descriptor */
  sint32_T s32FileId;
  /* write operation success */
  sint32_T s32Success;

  /* allocate memory for source array */
  if ( PRJxSET_bGetVerboseMode() ) {
    fprintf(stdout, "Ask for source array memory.\n");
  } 
  pcDataSource = (char *) malloc(u32ArraySize);

  /* check if the allocation failed */ 
  if ( pcDataSource  == NULL ) {	
    /* we through an exception */
    char aszErrorMsg[] = "Can't allocate enough memory for pcDataSource"; 
    EX_vSetError(aszErrorMsg, __FILE__, __LINE__);
    return(0); 
  }

  /* fill the source array */
  if ( PRJxSET_bGetVerboseMode() ) {
    fprintf(stdout, "Fill the source array with a saw tooth.\n");
  } 
  ME_vGenUint8Array(u32ArraySize, (uint8_T *) pcDataSource); 

  /* open the file */
  if ( PRJxSET_bGetVerboseMode() ) {
    fprintf(stdout, "Open file: %s\n", pszFileName);
  } 
  /* we open the output file (or overwrite it if it exists) */
  s32FileId = open(pszFileName, (O_RDWR | O_TRUNC | O_CREAT) ); 
  
  /* we check if the opening was successful */
  if ( s32FileId == -1 ) {
    /* we through an exception */
    char aszErrorMsg[] = "Can't open pszFileName"; 
    EX_vSetError(aszErrorMsg, __FILE__, __LINE__);
    return(0); 
  } 
  
  /* start the time measurement */
  if ( PRJxSET_bGetVerboseMode() ) {
    fprintf(stdout, "Start the file write process.\n");
  }
  ME_vStartTimer();

  /* write the output stream */
  s32Success = write(s32FileId, pcDataSource, u32ArraySize); 

  /* we check if the writing of the data was successful */ 
  if ( s32Success == -1 ) { 
    /* we through an exception */
    char aszErrorMsg[] = "We couldn't write the data information"; 
    EX_vSetError(aszErrorMsg, __FILE__, __LINE__);
  }

  /* we close the output stream */
  close(s32FileId);

  /* stop the time measurement */ 
  ME_vStopTimer(); 

  /* free the allocated memory again */
  if ( PRJxSET_bGetVerboseMode() ) {
    fprintf(stdout, "Give source array memory free.\n");
  }
  free(pcDataSource); 

  /* we set the correct mode (read, write for user, group and others),
   * this is needed for normal users to access the file during read
   * test */
  s32Success = chmod(pszFileName, 0000666);
  
  /* we check if changing of the mode was successful */ 
  if ( s32Success == -1 ) { 
    /* we through an exception */
    char aszErrorMsg[] = "We couldn't set the file mode"; 
    EX_vSetError(aszErrorMsg, __FILE__, __LINE__);
  } 
  
  return( ME_u64GetTimeDiff() ); 
}


/* -------------------------------------------------------------------- */ 
/* Read a file from the hard disc */
/* -------------------------------------------------------------------- */ 
static uint64_T u64ReadFile(char * pszFileName) {

  /* the data destination array size */
  uint32_T u32ArraySize; 
  /* the data destination array */
  char * pcDataDest;
  /* file descriptor */
  sint32_T s32FileId;
  /* write operation success */
  sint32_T s32Success;

  /* open the file */
  if ( PRJxSET_bGetVerboseMode() ) {
    fprintf(stdout, "Open file: %s\n", pszFileName);
  } 
   /* we open the output file (read only) */
  s32FileId = open(pszFileName, O_RDONLY );

  /* we check if the opening was successful */
  if ( s32FileId == -1 ) {
    /* we through an exception */
    char aszErrorMsg[] = "Can't open pszFileName"; 
    EX_vSetError(aszErrorMsg, __FILE__, __LINE__);
    return(0); 
  }

  /* get size of file */ 
  u32ArraySize = lseek(s32FileId, 0L, SEEK_END);
  lseek(s32FileId, 0L, SEEK_SET);

  /* for debugging only
  fprintf(stdout, "Array size to read: %d", u32ArraySize);
  */

  /* allocate memory for destination array */
  if ( PRJxSET_bGetVerboseMode() ) {
    fprintf(stdout, "Ask for destination array memory.\n");
  } 
  pcDataDest = (char *) malloc(u32ArraySize); 
  
  /* check if the allocation failed */ 
  if ( pcDataDest  == NULL ) {	
    /* we through an exception */
    char aszErrorMsg[] = "Can't allocate enough memory for pcDataDest"; 
    EX_vSetError(aszErrorMsg, __FILE__, __LINE__);
    return(0); 
  }	

  /* start the time measurement */
  if ( PRJxSET_bGetVerboseMode() ) {  
    fprintf(stdout, "Start the file read process.\n");
  }
  ME_vStartTimer();

  /* read the input stream */
  s32Success = read(s32FileId, pcDataDest, u32ArraySize); 

  /* we check if the reading of the data was successful */ 
  if ( s32Success == -1 ) { 
    /* we through an exception */
    char aszErrorMsg[] = "We couldn't read the data information"; 
    EX_vSetError(aszErrorMsg, __FILE__, __LINE__);
  } 

  /* we close the input stream */
  close(s32FileId);

  /* stop the time measurement */
  ME_vStopTimer(); 

  /* free the allocated memory again */
  if ( PRJxSET_bGetVerboseMode() ) {
    fprintf(stdout, "Give destination array memory free.\n");
  }
  free(pcDataDest); 

  return( ME_u64GetTimeDiff() ); 
} 


/* -------------------------------------------------------------------- */ 
/* Print out the time value to the command line */
/* -------------------------------------------------------------------- */ 
static void vCmdLineOut(char * pszMessage, uint64_T u64TimeDiff) {
  char aszOutput[80u]; 

  /* compose the output string */
  snprintf(aszOutput, 80u, "%s %lldus.", pszMessage, u64TimeDiff); 

  /* write it to stdout and to the logging file */
  ME_vAddLine(aszOutput);
}


/* -------------------------------------------------------------------- */ 
/* Print out the time value including MIN/MAX value to the command line */
/* -------------------------------------------------------------------- */ 
static void vCmdLineOutArray(char * pszMessage, uint64_T * pu64TimeDiffs, 
    uint32_T u32CntOfBytes) {
  
  char aszOutput[80u]; 

  uint64_T u64TimeDiffMax = 0;
  uint64_T u64TimeDiffMin = 0xFFFFFFFFFFFFFFFF;
  uint64_T u64TimeDiffMean = 0;
  uint64_T u64TimeDiffTemp;
  uint32_T u32Idx;
  float r32BytesPerMicroSecond;
  float r32KiloBytesPerSecond;

  for (u32Idx = 0; u32Idx < mg_u32MeasRepetitions; u32Idx++) {
    u64TimeDiffTemp = *pu64TimeDiffs++;

    if (u64TimeDiffTemp > u64TimeDiffMax) {
      u64TimeDiffMax = u64TimeDiffTemp;
    }

    if (u64TimeDiffTemp < u64TimeDiffMin) {
      u64TimeDiffMin = u64TimeDiffTemp;
    }

    /* sum up for the mean value */
    u64TimeDiffMean += u64TimeDiffTemp;
  }

  /* calculate the mean value */
  u64TimeDiffMean /= mg_u32MeasRepetitions;

  /* try to remove the cache size dependency (risk: what is the influence of the overhead?) */
  r32BytesPerMicroSecond = ((float) u32CntOfBytes) / ((float) u64TimeDiffMean);

  /* us in s is 1e6 / bytes in kilo bytes is 1000 -> multiple with 1000 */
  r32KiloBytesPerSecond = r32BytesPerMicroSecond * 1000.0f;

  /* put out all values */
  snprintf(aszOutput, 80u, "%s %lldus.", pszMessage, u64TimeDiffMean); 
  ME_vAddLine(aszOutput);
  snprintf(aszOutput, 80u, "(This are: %.3f kBytes per second.)", r32KiloBytesPerSecond); 
  ME_vAddLine(aszOutput); 
  snprintf(aszOutput, 80u, "The maximum value was: %lldus.", u64TimeDiffMax); 
  ME_vAddLine(aszOutput); 
  snprintf(aszOutput, 80u, "The minimum value was: %lldus.\n", u64TimeDiffMin); 
  ME_vAddLine(aszOutput); 
}

