/***********************************************************************
 * ME_Time.c
 *
 * Measure the time consumption for a function in microseconds
 *
 * The time measurement is based on ideas of Alexander Saprykin
 * published on:
 * https://stackoverflow.com/questions/361363
 * /how-to-measure-time-in-milliseconds-using-ansi-c
 *
 ***********************************************************************
 * License
 * (C) Copyright 2022 Falk Richter, falk.richter at yandex.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 ***********************************************************************/ 

/* needed for fprintf */
#include <stdio.h>

#if defined __linux__ && !defined __USE_POSIX199309 
  #define __USE_POSIX199309
#endif

#include <time.h>
#include <sys/time.h>
#include <unistd.h>

/* error handling */
#include "EX_Error.h"
/* the own header */
#include "ME_Time.h"

static uint64_T u64GetClkTime(void);
static uint64_T mg_u64StartValue;
static uint64_T mg_u64EndValue;

#if defined CLOCK_REALTIME
  /* variable is only needed if we have (at least) a realtime clock */
  static clockid_t mg_clk_id = CLOCK_REALTIME;
#endif

enum {eIdle, eStarted, eStopped} mg_enState;

/* -------------------------------------------------------------------- */
/* initialize the timer module */
/* -------------------------------------------------------------------- */
void ME_vTimerInit(void) {
/* CLOCK_REALTIME is also needed otherwise mg_clk_id doesn't exist */
#if defined _SC_MONOTONIC_CLOCK && defined CLOCK_MONOTONIC && defined CLOCK_REALTIME 
  if ( sysconf (_SC_MONOTONIC_CLOCK) != -1 ) {
    /* A monotonic clock is present we take this one */ 
    mg_clk_id = CLOCK_MONOTONIC;
    fprintf(stdout,"(We are using clock_gettime with a monotonic clock.)");
  } else {
    fprintf(stdout,"(We are using clock_gettime with a real time clock.)"); 
  }
#else
  #if defined CLOCK_REALTIME 
    fprintf(stdout,"(We are using clock_gettime with a real time clock.)"); 
  #else 
    fprintf(stdout,"(We are using gettimeofday for time measurement.)");
  #endif
#endif
  fprintf(stdout, "\n\n");

  mg_u64StartValue = 0;
  mg_u64EndValue = 0;
  mg_enState = eIdle;
} 


/* -------------------------------------------------------------------- */ 
/* function starts the time measurement */
/* -------------------------------------------------------------------- */
void ME_vStartTimer(void) {
  /* fetch the current time */
  mg_u64StartValue = u64GetClkTime();
  /* check if we got a time stamp greater than zero */
  if ( mg_u64StartValue > 0 ) {
    /* we set the current state */
    mg_enState = eStarted;
  } 
} 


/* -------------------------------------------------------------------- */ 
/* function stops the time measurement */
/* -------------------------------------------------------------------- */ 
void ME_vStopTimer(void) {
  /* check if a measurement is running */
  if ( mg_enState != eStarted ) {
    /* we through an exception */
    char aszErrorMsg[] = "No time measurement was started"; 
    EX_vSetError(aszErrorMsg, __FILE__, __LINE__);
  }

  /* fetch the current time */
  mg_u64EndValue = u64GetClkTime();
  /* check if we got a time stamp greater than zero */
  if ( mg_u64EndValue > 0 ) {
    /* we set the current state */
    mg_enState = eStopped;
  } 
}


/* -------------------------------------------------------------------- */ 
/* function calculates the time difference and returns the result */
/* -------------------------------------------------------------------- */ 
uint64_T ME_u64GetTimeDiff(void) {
  /* check if a measurement is finished */
  if ( mg_enState != eStopped ) {
    /* we through an exception */	
    char aszErrorMsg[] = "No time measurement was stopped"; 
    EX_vSetError(aszErrorMsg, __FILE__, __LINE__);
  } 

  /* return the difference */
  return(mg_u64EndValue - mg_u64StartValue);
}


/* -------------------------------------------------------------------- */ 
/* function returns a time stamp in micro seconds */
/* -------------------------------------------------------------------- */ 
uint64_T u64GetClkTime(void) {

#if defined CLOCK_REALTIME 
  /* we are using clock_gettime if a monotonic clock is available */
  struct timespec strTs;

  /* try to get the time */
  if ( clock_gettime(mg_clk_id, &strTs) == 0 ) {
	/* clock_gettime delivers seconds and nano seconds */
	return((uint64_T) (strTs.tv_sec * 1000000u + strTs.tv_nsec / 1000u));
  }
  /* if clock_gettime fails we return zero */
  return 0;
#else
  /* this is the older version: no monotonic clock, the result
  could be wrong if anybody adjusts the time during the start and
  the end time stamp */

  struct timeval strTv;

  if ( gettimeofday(&strTv, NULL) == 0 ) {
    /* gettimeofday delivers seconds and micro seconds */
    return((uint64_T) (strTv.tv_sec * 1000000u + strTv.tv_usec));
  }
  /* if gettimeofday fails we return zero */ 
  return 0;
#endif 
}

