/***********************************************************************
 * PerfTest.c
 *
 * Simple program to test the performance of my old computers  
 * (cache, memory and file access)
 *
 * The time measurement is based on ideas of Alexander Saprykin
 * published on:
 * https://stackoverflow.com/questions/361363
 * /how-to-measure-time-in-milliseconds-using-ansi-c
 *
 ***********************************************************************
 * License
 * (C) Copyright 2020 Falk Richter, falk.richter at yandex.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 ***********************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "PRJ_Types.h"

#include "PRJ_Settings.h" 
#include "EX_Error.h"
#include "ME_Measure.h"
#include "ME_LogFile.h"


#define ERROR_MSG_LEN 80u
static char mg_aszErrorMsg[ERROR_MSG_LEN];

/* a short help */
static void vUsage(void);

/* put out a message in case of an error during the argument parse process */
static void vParseError(void);

/* check for an overtaken value after an argument switch */
static bool_T bCheckForValue( sint32_T s32Index, sint32_T s32Argc, char* pszArgv[] );

/* function to parse the main arguments */ 
static bool_T bGetOpt(char * pszArgument, char * pszOption);

/* the software version */
static const char aszSwVersionNumber[] = "0.4.1";


/* -------------------------------------------------------------------- */ 
/* the main program  */
/* -------------------------------------------------------------------- */ 
int main(int argc, char* argv[]) {
  sint32_T s32i; /* loop index */
  sint32_T s32ReturnVal = EXIT_SUCCESS; /* we hope the best */

  /* Print the program header at the beginning */ 
  fprintf(stdout, "\n");
  fprintf(stdout, "PerfTest\n");
  fprintf(stdout, "--------\n");
  fprintf(stdout, "is a simple program to test the performance of my historical\n");
  fprintf(stdout, "486 / 586 computers.\n"); 
  fprintf(stdout, "\n");
  fprintf(stdout, "The program stands under the terms of the GPL3 licence\n"); 
  fprintf(stdout, "Copyright 2022 Falk Richter <falk.richter at yandex.com>, version: %s\n", aszSwVersionNumber);
  fprintf(stdout, "\n");

  /* Initialize the project settings module */
  PRJxSET_vInit();

  /* we start the parsing at argv[1] because we aren't interested in the */ 
  /* path name of the program (argument argv[0]) */
  for ( s32i = 1; s32i < argc; ) {

    /* the types of tests */
    if ( bGetOpt(argv[s32i], "t") == TRUE ) {
      /* we test if there a further argument with the value */
      if ( bCheckForValue(s32i, argc, argv) ) {
	/* the next argument must be the value for "-t" so we can skip */
	/* to parse it - that's the reason of the increment of i */
	if ( PRJxSET_bSetTestTypes((uint32_T) atoi(argv[++s32i])) == FALSE ) {
	  /* value for test types isn't valid */
	  snprintf(mg_aszErrorMsg, ERROR_MSG_LEN, "Value for -t isn't valid: %s", argv[s32i]); 
	  vParseError();
	}
      }
    } 
    /* the size of the cache */
    else if ( bGetOpt(argv[s32i], "c") == TRUE ) {
      /* we test if there a further argument with the value */
      if ( bCheckForValue(s32i, argc, argv) ) {
	/* the next argument must be the value for "-c" so we can skip */
	/* to parse it - that's the reason of the increment of i */
	if ( PRJxSET_bSetCacheSize((uint32_T) atoi(argv[++s32i])) == FALSE ) {
	  /* value for cache size isn't valid */
	  snprintf(mg_aszErrorMsg, ERROR_MSG_LEN, "Value for -c isn't valid: %s", argv[s32i]); 
	  vParseError();
	}
      }
    }    /* verbose argument */
    else if ( bGetOpt(argv[s32i], "v") == TRUE ) {	
      /* verbose option has no arguments */
      PRJxSET_bSetVerboseMode(TRUE);
    } 
    /* help argument */
    else if ( bGetOpt(argv[s32i], "h") == TRUE ) {
      /* print a short help */
      vUsage();
      return(EXIT_SUCCESS);
    } else {
      /* an argument error occurred */

      /* compose an error message */ 
      snprintf(mg_aszErrorMsg, ERROR_MSG_LEN, "Unknown option: %s", argv[s32i]); 
      vParseError(); 
    }
    /* prepare the next index */
    s32i++; 
  }

  /* open the logging file */
  if ( ME_bStartLogFile() == FALSE ) {
    /* an error occurred, prints the error message */
    EX_vPrintReason();
    return(EXIT_FAILURE);
  }

  /* initialize the measurement */
  ME_vMeasInit(); 

  /* start the measurement */
  if ( ME_bMeasStart() == FALSE ) {
    /* an error occurred, prints the error message */
    EX_vPrintReason();
    s32ReturnVal = EXIT_FAILURE;
  } 

  /* close the logging file again */
  ME_vCloseLogFile();

  return(s32ReturnVal); 
}


/* -------------------------------------------------------------------- */ 
/* in case of a parse error we left the program  */
/* -------------------------------------------------------------------- */ 
static void vParseError(void) {
  /* put out the reason of the error */
  fprintf(stdout, "Error: %s\n", mg_aszErrorMsg);
  fprintf(stdout, "******\n");
  fprintf(stdout, "\n");

  /* we print additionally the help */
  vUsage();

  exit(EXIT_FAILURE);
}


/* -------------------------------------------------------------------- */ 
/* check for an overtaken value after an argument switch */
/* -------------------------------------------------------------------- */ 
static bool_T bCheckForValue( sint32_T s32Index, sint32_T s32Argc, char * pszArgv[] ) {
  /* return value of this function -> we hope the best */
  bool_T bReturnVal = TRUE;
  /* make a copy of the option which calls bCheckForValue */
  char * pszOption = pszArgv[s32Index];

  /* we check if there a further argument with a value for this option switch */
  s32Index++;

  /* check if pszOption is not already the last value in the argument array */
  if (s32Index < s32Argc) {
    /* the next argument must be a value and not the (next) option switch */
    char * pszArgument = pszArgv[s32Index]; 

    /* the "-" at the first position is the token for an option switch */
    /* if we found a leading "-" the return value of the compare function is zero */
    if ( pszArgument[0] == '-' ) {
      /* this is the next option and is not correct */
      bReturnVal = FALSE;
    }	
  } else {
    /* pszOption is the last field / value in the argument array -> no value was handed over */
    bReturnVal = FALSE;
  }
  /* check for errors */
  if (bReturnVal == FALSE) {
    /* compose an error message */ 
    snprintf(mg_aszErrorMsg, ERROR_MSG_LEN, "No value was overtaken for the option: %s", pszOption); 
    vParseError(); 
  } 

  return(bReturnVal); 
}


/* -------------------------------------------------------------------- */ 
/* a short help function  */
/* -------------------------------------------------------------------- */ 
static void vUsage(void) {
  /* the short user help */
  fprintf(stdout, "Usage:\n");
  fprintf(stdout, "------\n");
  fprintf(stdout, "-t <Testnumber(s)> see also \"Kind of tests\" (default: 7 (Test 1 + 2 + 4))\n");
  fprintf(stdout, "-c <Cache size in kBytes> (default: 256, valid are: 64, 128, 256, 512 and 1024)\n"); 
  fprintf(stdout, "-v verbose mode\n");
  fprintf(stdout, "-h prints this help\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "Kind of tests:\n"); 
  fprintf(stdout, "1 - memory test\n");
  fprintf(stdout, "2 - cache test\n");
  fprintf(stdout, "4 - file write / read test\n");
  fprintf(stdout, "\n");
}


/* -------------------------------------------------------------------- */ 
/* function to parse the main arguments  */
/* -------------------------------------------------------------------- */ 
static bool_T bGetOpt(char * pszArgument, char * pszOption) {

  /* returns true if the argument starts with a minus sign and */
  /* is equal to the handed over option */
  if ( (pszArgument[0] == '-') && (pszArgument[1] == pszOption[0]) ) {
    return(TRUE);
  } else {
    return(FALSE);
  }
}

